import 'dart:io';
import 'dart:mirrors';
import 'package:easy_tui/cli/io/out/terminal_out.dart';
import 'package:easy_tui/tui/layouts.dart';
import 'package:easy_tui/tui/menu_modifiers.dart';
import 'package:easy_tui/tui/text_dialogue.dart';
import 'package:easy_tui/tui/text_option_ui.dart';
import 'package:easy_tui/tui/text_ui.dart';

/// A subtype of `TextUI` that accepts variable input from a user.
class TextFormUI extends TextUI<List<String>> {
  /*
	 * FIELDS
	 */

  static const String _confirmExit =
      'Is "${TextUI.EXIT}" your input ("${TextDialogue.CHOICE_YES}"), '
      'or do you wish to quit data entry ("${TextDialogue.CHOICE_NO}")?';

  /*
	 * CONSTRUCTOR
	 */

  TextFormUI({List<MenuModifiers> modifiers, Layouts layout = Layouts.VERTICAL})
      : super(modifiers: modifiers, layout: layout);

  /*
	 * OPERATIONS
	 */

  ///	Displays a custom error using [answers] and a reported [invalidIndex].
  ///
  ///	[<T>] is the desired type of the invalid entry.
  ///	The `dart:mirrors` library is used to determine the type of the example.
  static void displayInvalidEntry<T>(
      List<String> answers, int invalidIndex, T example) {
    print('One of your entries was invalid. See below:'
        '\n"${answers[invalidIndex]}" is not a valid `${reflect(example).type.simpleName}`');
    TerminalOut.enterToClear();
  }

  /// Ask the user to enter input for a series of prompts specified in [this.lines].
  @override
  List<String> present({String preface, String prepend}) {
    ///		Labelled "frozen" because any concurrent changes are not reflected during [present()]
    final List<String> frozenQuestions =
        this.toStringIterative(prepend: prepend);
    final List<String> answers = new List<String>(frozenQuestions.length);
    final String interfaceDirections = this.promptModifierAdjustments();

    if (preface == null) {
      TerminalOut.clear();
    } else {
      TerminalOut.enterToClear(preface);
    }

    for (int i = 0; i < frozenQuestions.length; i++) {
      print(
          "${frozenQuestions[i]}\n$interfaceDirections${TextDialogue.PROMPT}");
      final String input = stdin.readLineSync();

      ///			`true` if the user signifies that they wish to leave
      if ((input == TextUI.EXIT) && (TextDialogue.yes(_confirmExit))) {
        return null;
      } else {
        answers[i] = input;
      }
      TerminalOut.clear();
    }
    TerminalOut.enterToClear();

//		Give the user a chance to correct any mistakes that they may have made
    return TextFormUI.reviseAnswers(
        prefaces: this.toStringIterative(
//						Regenerate the questions without the number-bullets
//						This is to prevent a number-bullet duplication bug
            prepend: prepend,
            numberBullets: false),
        revisionData: answers);
  }

  /// Prompt user with [prefaces] to change [revisionData].
  static List<String> reviseAnswers(
      {List<String> prefaces, List<String> revisionData}) {
    const String continueChanging = "Would you like to change any of them?";
    while (true) {
      final TextOptionUI selectData =
          new TextOptionUI(modifiers: <MenuModifiers>[MenuModifiers.NO_EXIT]);
      print("Your answers:\n=========================");
      for (int i = 0; i < revisionData.length; ++i) {
        selectData.lines.add(

            ///						[TextDialogue.PROMPT] has a "\n", so no "\n" is needed here
            "${prefaces[i]}${TextDialogue.PROMPT} ${revisionData[i]}\n");
        print(selectData.lines.last);
      }

      if (TextDialogue.yes(continueChanging)) {
        ///				[present()] is a blocking call
        final int revisionIndex = selectData.present(
                preface:
                    "Enter the number of which you would like to change.") -
            1;

        TerminalOut.clear();

        print("${prefaces[revisionIndex]}"
//						Give the previous answer, for reference
            "\nPrevious answer: ${revisionData[revisionIndex]}"
            "\n${TextDialogue.PROMPT}");

        final String input = stdin.readLineSync();
        if (input != "") {
          revisionData[revisionIndex] = input;
        }
      } else {
        break;
      }
      TerminalOut.enterToClear();
    }
    return revisionData;
  }
}
